<?php

/**
 * @file
 * Connector to Memsource Service.
 */

/**
 * Memsource connector class.
 */
class MemsourceConnector {

  private $translator;

  private $serviceUrl;

  private $userName;

  private $password;

  private $token;

  private $moduleVersion;

  /**
   * MemsourceConnector Constructor.
   *
   * @param string $service_url
   *   Memsource API URL.
   * @param string $user_name
   *   Memsource Cloud user name.
   * @param string $password
   *   Memsource Cloud password.
   * @param null|object $translator
   *   TMGMT Translator.
   */
  public function __construct($service_url, $user_name, $password, $translator) {
    $this->serviceUrl = $service_url;
    $this->userName = $user_name;
    $this->password = $password;
    $this->translator = $translator;
    $this->moduleVersion = null;
  }

  /**
   * {@inheritdoc}
   */
  public function getSupportedRemoteLanguages() {
    $supported_remote_languages = [];

    if (empty($this->userName) || empty($this->password)) {
      return $supported_remote_languages;
    }

    try {
      $supported_languages = $this->sendApiRequest('/api2/v1/languages');
      foreach ($supported_languages['languages'] as $language) {
        $supported_remote_languages[$language['code']] = $language['code'];
      }
    }
    catch (\Exception $e) {
      // Ignore exception, nothing we can do.
    }

    asort($supported_remote_languages);

    return $supported_remote_languages;
  }

  /**
   * {@inheritdoc}
   */
  public function requestJobItemsTranslation(array $job_items) {
    /** @var TMGMTJob $job */
    $job = reset($job_items)->getJob();
    $due_date_array = $job->getSetting('due_date');
    $date_time = new DateTime();
    $date_time->setDate($due_date_array['year'], $due_date_array['month'], $due_date_array['day']);
    $due_date = $date_time->format('Y-m-d');
    try {
      $project_id = $this->newTranslationProject($job, $due_date);
      $job->addMessage('Created a new project in Memsource Cloud with the id: @id', ['@id' => $project_id], 'debug');

      /** @var TMGMTJobItem $job_item */
      foreach ($job_items as $job_item) {
        $job_part_id = $this->sendFiles($job_item, $project_id, $due_date);
        /** @var TMGMTRemote $remote_mapping */
        $remote_mapping = entity_create('tmgmt_remote', [
          'tjid' => $job->tjid,
          'tjiid' => $job_item->tjiid,
          'remote_identifier_1' => 'tmgmt_memsource',
          'remote_identifier_2' => $project_id,
          'remote_identifier_3' => $job_part_id,
          'remote_data' => [
            'FileStateVersion' => 1,
            'TmsState' => 'TranslatableSource',
            'RequiredBy' => $due_date,
          ],
        ]);
        $remote_mapping->save();
      }
      $this->assignMemsourceProviders($project_id, $job);
    }
    catch (TMGMTException $e) {
      $job->rejected('Job has been rejected with following error: @error',
        ['@error' => $e->getMessage()], 'error');
      if (isset($remote_mapping)) {
        $remote_mapping->delete();
      }
    }
    return $job;
  }

  /**
   * Performs a login to Memsource Cloud.
   *
   * @return bool
   *   A success or failure.
   */
  public function loginToMemsource() {
    $params = [
      'userName' => $this->userName,
      'password' => $this->password,
    ];
    try {
      $result = $this->request('/api2/v1/auth/login', 'POST', $params);
      if ($result['token']) {
        // Store the token.
        $this->storeToken($result['token']);
        return TRUE;
      }
      throw new TMGMTException('Token not found in response from Memsource!');
    }
    catch (TMGMTException $ex) {
      if (variable_get('tmgmt_memsource_debug', FALSE)) {
        watchdog('tmgmt_memsource', 'Unable to log in to Memsource API: @error', ['@error' => $ex->getMessage()], WATCHDOG_DEBUG);
      }
    }
    return FALSE;
  }

  /**
   * Stores a Memsource API token.
   *
   * @param string $token
   *   Token.
   */
  public function storeToken($token) {
    $this->token = $token;
  }

  /**
   * Returns a Memsource API token.
   *
   * @return string
   *   Token.
   */
  public function getToken() {
    return $this->token;
  }

  /**
   * Sends a request to the Memsource API and refreshes the token if necessary.
   *
   * @param string $path
   *   API path.
   * @param string $method
   *   (Optional) HTTP method.
   * @param array $params
   *   (Optional) API params.
   * @param bool $code
   *   (Optional) If true, return only the response HTTP status code.
   * @param bool $download
   *   (Optional) If true, return the response body as a downloaded content.
   * @param string $body
   *   (Optional) An optional request body.
   *
   * @return array|int|null
   *   Result of the API request.
   *
   * @throws TMGMTException
   */
  public function sendApiRequest($path, $method = 'GET', array $params = [], $code = FALSE, $download = FALSE, $body = NULL) {
    $result = NULL;
    $params['token'] = $this->getToken();
    try {
      $result = $this->request($path, $method, $params, $code, $download, $body);
    }
    catch (TMGMTException $ex) {
      if ($ex->getCode() === 401) {
        // Token is invalid, try to re-login.
        $this->loginToMemsource();
        $params['token'] = $this->getToken();
        $result = $this->request($path, $method, $params, $code, $download, $body);
      }
      else {
        throw $ex;
      }
    }
    return $result;
  }

  /**
   * Does a request to Memsource API.
   *
   * @param string $path
   *   Resource path.
   * @param string $method
   *   (Optional) HTTP method (GET, POST...). By default uses GET method.
   * @param array $params
   *   (Optional) Form parameters to send to memsource service.
   * @param bool $code
   *   (Optional) If we want to return the status code of the call. FALSE by
   *   default.
   * @param bool $download
   *   (Optional) If we expect resource to be downloaded. FALSE by default.
   * @param string $body
   *   (Optional) Body of the POST request. NULL by
   *   default.
   *
   * @return array|int|null
   *   Response array from memsource.
   *
   * @throws TMGMTException
   */
  public function request($path, $method = 'GET', array $params = [], $code = FALSE, $download = FALSE, $body = NULL) {
    $options = ['method' => $method, 'headers' => []];
    $query = [];

    if (isset($params['token'])) {
      $options['headers'] = ['Authorization' => 'ApiToken ' . $params['token']];
      unset($params['token']);
    }

    if ($body) {
      $options['data'] = $body;
      if (!empty($params)) {
        $options['headers'] += $params;
      }
    }
    elseif ($method === 'GET') {
      $query = $params;
    }
    else {
      $options['data'] = json_encode($params);
      $options['headers'] += ['Content-Type' => 'application/json'];
    }

    $url = $this->updateServiceUrl($this->serviceUrl) . $path . '?' . http_build_query($query);

    $log = [
      '%method' => $method,
      '%url' => $url,
      '%headers' => isset($options['headers']) ? json_encode($options['headers']) : '',
      '%data' => isset($options['data']) ? $options['data'] : '',
    ];

    $response = drupal_http_request($url, $options);

    if (!in_array($response->code, ['200', '201'], TRUE)) {
      watchdog('tmgmt_memsource', "\n=> REQUEST:\n%method %url\nHeaders: %headers\nData: %data\n\n=> RESPONSE:\nCode: %code\nData: %response\n",
        $log + [
          '%response' => $response->data,
          '%code' => $response->code,
        ], WATCHDOG_ERROR);
      if ($code) {
        return $response->code;
      }
      throw new TMGMTException('Unable to connect to Memsource service due to following error: @error at @url',
        ['@error' => $response->data, '@url' => $url], $response->code);
    }

    watchdog('tmgmt_memsource', "\n=> REQUEST:\n%method %url\nHeaders: %headers\nData: %data\n\n=> RESPONSE:\nCode: %code\nData: %response\n",
      $log + [
        '%response' => $response->data,
        '%code' => $response->code,
      ], WATCHDOG_DEBUG);

    if ($code) {
      return $response->code;
    }

    // If we are expecting a download, just return received data.
    if ($download) {
      return $response->data;
    }

    return json_decode($response->data, TRUE);
  }

  /**
   * Modify service URL due to the Memsource API migration (if necessary).
   *
   * 'https://qa.memsource.com/web/api' -> 'https://qa.memsource.com/web'.
   *
   * @param string $url
   *   Service URL.
   *
   * @return string
   *   Updated Memsource URL.
   */
  private function updateServiceUrl($url) {
    $pos = strpos($url, '/web/');

    if ($pos !== FALSE) {
      $url = substr($url, 0, $pos + 4);
      // $this->translator->setSetting('service_url', $url);.
    }

    return $url;
  }

  /**
   * Creates new translation project at Memsource Cloud.
   *
   * @param TMGMTJob $job
   *   The job.
   * @param string $due_date
   *   The date by when the translation is required.
   *
   * @return int
   *   Memsource project id.
   *
   * @throws TMGMTException
   */
  public function newTranslationProject(TMGMTJob $job, $due_date) {
    $template_id = $job->getSetting('project_template');

    $params = [
      'name' => $job->label ?: 'Drupal TMGMT project ' . $job->tjid,
      'sourceLang' => $this->getMappedLanguage($job->source_language),
      'targetLangs' => [$this->getMappedLanguage($job->target_language)],
    ];

    if (strlen($due_date) > 0) {
      $params['dateDue'] = $due_date . 'T00:00:00Z';
    }

    if ($template_id === '0' || $template_id === NULL) {
      $result = $this->sendApiRequest('/api2/v1/projects', 'POST', $params);
    }
    else {
      $result = $this->sendApiRequest("/api2/v2/projects/applyTemplate/$template_id", 'POST', $params);
    }

    return $result['id'];
  }

  /**
   * Assign providers defined in project template.
   *
   * @param string $project_id
   *   Project UID.
   * @param TMGMTJob $job
   *   The job.
   */
  public function assignMemsourceProviders($project_id, TMGMTJob $job) {
    $template_id = $job->getSetting('project_template');
    if ($template_id !== NULL && $template_id !== '0') {
      $this->sendApiRequest(
        "/api2/v1/projects/$project_id/applyTemplate/$template_id/assignProviders",
        'POST'
      );
    }
  }

  /**
   * Send the files to Memsource Cloud.
   *
   * @param TMGMTJobItem $job_item
   *   The Job.
   * @param int $project_id
   *   Thebigword project id.
   * @param string $due_date
   *   The date by when the translation is required.
   *
   * @return int
   *   Memsource jobPartId.
   */
  public function sendFiles(TMGMTJobItem $job_item, $project_id, $due_date) {
    // Prepare parameters for Job API.
    $xliff_converter = tmgmt_file_format_controller('xlf');

    $job_item_id = $job_item->tjiid;
    $target_language = $this->getMappedLanguage($job_item->getJob()->target_language);
    $conditions = ['tjiid' => ['value' => $job_item_id]];
    $xliff = $xliff_converter->export($job_item->getJob(), $conditions);
    $name = "JobID_{$job_item->tjid}_JobItemID_{$job_item_id}_{$job_item->getJob()->source_language}_{$target_language}";

    $job_part_id = $this->createJob($project_id, $target_language, $due_date, $xliff, $name);

    return $job_part_id;
  }

  /**
   * Creates a new job at Memsource Cloud.
   *
   * @param string $project_id
   *   Project ID.
   * @param string $target_language
   *   Target language code.
   * @param string $due_date
   *   Job due date.
   * @param string $xliff
   *   XLIFF file.
   * @param string $name
   *   File name.
   *
   * @return string
   *   Memsource uuid of the resource.
   *
   * @throws TMGMTException
   */
  public function createJob($project_id, $target_language, $due_date, $xliff, $name) {
    $params = [
      'targetLangs' => [$this->getMappedLanguage($target_language)],
      'useProjectFileImportSettings' => 'true',
      'sourceData' => [
        'clientType' => 'DRUPAL',
        'clientVersion' => $this->getMemsourceModuleVersion(),
        'hostVersion' => defined('VERSION') ? VERSION : '7',
      ],
    ];

    if (strlen($due_date) > 0) {
      $params['due'] = $due_date . 'T00:00:00Z';
    }

    $headers = [
      'Content-Disposition' => "filename*=UTF-8''$name.xliff",
      'Memsource' => json_encode($params),
    ];

    $result = $this->sendApiRequest("/api2/v1/projects/$project_id/jobs", 'POST', $headers, FALSE, FALSE, $xliff);

    return $result['jobs'][0]['uid'];
  }

  /**
   * Fetches translations for job items of a given job.
   *
   * @param TMGMTJob $job
   *   A job containing job items that translations will be fetched for.
   *
   * @return array
   *   Array containing a containing the number of items translated and the
   *   number that has not been translated yet.
   */
  public function fetchTranslatedFiles(TMGMTJob $job) {
    $translated = 0;
    $errors = [];

    try {
      /** @var TMGMTJobItem $job_item */
      foreach ($job->getItems() as $job_item) {
        /** @var TMGMTRemoteController $remote_mapping_controller */
        $remote_mapping_controller = entity_get_controller('tmgmt_remote');
        $mappings = $remote_mapping_controller->loadByLocalData($job_item->tjid, $job_item->tjiid);
        /** @var TMGMTRemote $mapping */
        foreach ($mappings as $mapping) {
          $job_uid = $mapping->remote_identifier_3;
          $project_id = $mapping->remote_identifier_2;
          $info = [];
          try {
            $info = $this->sendApiRequest("/api2/v1/projects/$project_id/jobs/$job_uid");
          }
          catch (TMGMTException $e) {
            $job->addMessage('Error fetching the job item: @job_item. Memsource job @job_uid not found: @error',
              [
                '@job_item' => $job_item->label(),
                '@job_uid' => $job_uid,
                '@error' => $e->getMessage(),
              ], 'error');
            $errors[] = 'Memsource job ' . $job_uid . ' not found, it was probably deleted.';
          }
          if (isset($info['status']) && $this->remoteTranslationCompleted($info['status'])) {
            try {
              $this->addFileDataToJob($job, $info['status'], $project_id, $job_uid);
              $translated++;
            }
            catch (TMGMTException $e) {
              $job->addMessage('Error fetching the job item: @job_item.', ['@job_item' => $job_item->label()], 'error');
              continue;
            }
          }
        }
      }
    }
    catch (TMGMTException $e) {
      watchdog('tmgmt_memsource', 'Could not pull translation resources: @error', ['@error' => $e->getMessage()], $severity = WATCHDOG_ERROR);
    }

    return [
      'translated' => $translated,
      'untranslated' => count($job->getItems()) - $translated,
      'errors' => $errors,
    ];
  }

  /**
   * Retrieve all the updates for all the job items in a translator.
   *
   * @param TMGMTJobItem $job_item
   *   The translator.
   *
   * @return int
   *   The number of updated items.
   */
  public function pullRemoteTranslation(TMGMTJobItem $job_item) {
    $job = $job_item->getJob();
    $this->translator = $job->getTranslator();
    /** @var TMGMTRemoteController $remote_mapping_controller */
    $remote_mapping_controller = entity_get_controller('tmgmt_remote');
    $remotes = $remote_mapping_controller->loadByLocalData($job->tjid, $job_item->tjiid);
    /** @var TMGMTRemote $remote */
    $remote = reset($remotes);
    $job_uid = $remote->remote_identifier_3;
    $project_id = $remote->remote_identifier_2;
    $info = $this->sendApiRequest("/api2/v1/projects/$project_id/jobs/$job_uid");
    if (isset($info['status']) && $this->remoteTranslationCompleted($info['status'])) {
      try {
        $this->addFileDataToJob($job, $info['status'], $project_id, $job_uid);
        return 1;
      }
      catch (TMGMTException $e) {
        $job->addMessage('Error fetching the job item: @job_item.', [
          '@job_item' => $remote->getJobItem()->label(),
        ], 'error');
      }
    }
    return 0;
  }

  /**
   * Retrieve the data of a file in a state.
   *
   * @param TMGMTJob $job
   *   The Job to which will be added the data.
   * @param string $state
   *   The state of the file.
   * @param int $project_id
   *   The project id.
   * @param string $job_uid
   *   The job ID.
   *
   * @throws TMGMTException
   */
  public function addFileDataToJob(TMGMTJob $job, $state, $project_id, $job_uid) {
    $data = $this->sendApiRequest("/api2/v1/projects/$project_id/jobs/$job_uid/targetFile", 'GET', [], FALSE, TRUE);
    $file_data = $this->parseTranslationData($data);
    $job->addTranslatedData($file_data, []);
    /** @var TMGMTRemoteController $remote_mapping_controller */
    $remote_mapping_controller = entity_get_controller('tmgmt_remote');
    $mappings = $remote_mapping_controller->loadByRemoteIdentifier('tmgmt_memsource', $project_id, $job_uid);
    /** @var TMGMTRemote $mapping */
    $mapping = reset($mappings);
    $mapping->removeRemoteData('TmsState');
    $mapping->addRemoteData('TmsState', $state);
    $mapping->save();
  }

  /**
   * Parses received translation from Memsource and returns unflatted data.
   *
   * @param string $data
   *   Base64 encode data, received from Memsource.
   *
   * @return array
   *   Unflatted data.
   */
  protected function parseTranslationData($data) {
    $xlf_converter = tmgmt_file_format_controller('xlf');
    /** @var TMGMTFileformatXLIFF $xlf_converter */
    $xlf_converter = new $xlf_converter();
    // Import data using XLIFF converter, passed content is not a file.
    return $xlf_converter->import($data, FALSE);
  }

  /**
   * Get mapped or original language.
   *
   * @param string $language
   *   Given language.
   *
   * @return string
   *   Mapped language.
   */
  public function getMappedLanguage($language) {
    $remote_languages_mappings = $this->translator->settings['remote_languages_mappings'];
    return (isset($remote_languages_mappings[$language]) && $remote_languages_mappings[$language]) ? $remote_languages_mappings[$language] : $language;
  }

  /**
   * Checks if the translation has one of the completed statuses.
   *
   * @param string $status
   *   Status code.
   *
   * @return bool
   *   True if completed.
   */
  public function remoteTranslationCompleted($status) {
    return in_array($status, ['COMPLETED_BY_LINGUIST', 'COMPLETED', 'DELIVERED']);
  }

  /**
   * Get version of Memsource plugin.
   *
   * @return string
   *   Current version.
   */
  private function getMemsourceModuleVersion() {
    if ($this->moduleVersion === NULL) {
      try {
        $info = drupal_parse_info_file(__DIR__ . DIRECTORY_SEPARATOR . 'tmgmt_memsource.info');
        $this->moduleVersion = (isset($info['version']) ? $info['version'] : '');
      }
      catch (\Exception $e) {
        $this->moduleVersion = '';
      }
    }
    return $this->moduleVersion;
  }

}
